from datetime import datetime
from pymarc import (Field, Subfield, JSONReader, Record)
from typing import List, Optional, Iterator
from rara_tools.normalizers.reader import SafeJSONReader

from rara_tools.normalizers.viaf import VIAFRecord, VIAFClient
from rara_tools.constants.normalizers import (
    DEFAULT_VIAF_FIELD, ALLOWED_VIAF_FIELDS, ALLOWED_VIAF_WIKILINK_LANGS,
    VIAF_SIMILARITY_THRESHOLD, VERIFY_VIAF_RECORD, MAX_VIAF_RECORDS_TO_VERIFY,
    EMPTY_INDICATORS
)
from glom import glom
import logging
import json

logger = logging.getLogger(__name__)


class RecordNormalizer:
    """
    Base class. For normalizing different record types corresponding classes have been created.
    By default existing record fields will not be changed, unless included in ALLOW_EDIT_FIELDS. If a field
    included in the normalization is not present, it will be added to the record. If under REPEATABLE_FIELDS.
    a new record field is added.

    Args:
        sierra_data: Optionally, can normalize records from SIERRA. Must be in specific format,
        e.g converted with SierraResponseConverter. examples at: tests/sierra/output
        entities: List of Full names (str). If included, will use NormLinker to match with normalized records on KATA elastic.
    """

    def __init__(self, linking_results: List[dict] = [], sierra_data: List[dict] = [],
                 ALLOW_EDIT_FIELDS: List[str] = ["667", "925"], REPEATABLE_FIELDS: List[str] = []):

        # Include, if will replace existing field
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        # include, if should be added alongside existing fields
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS
        self.records_extra_data = []
        self.records = self._setup_records(linking_results, sierra_data)
        self.sierra_data = sierra_data

    def _setup_records(self, linking_results: List[dict], sierra_data: List[dict]) -> JSONReader:
        """Setup initial MARC records and data.

        For linked entities:
            1. Try to get single linked normalized record from KATA elastic. If more than one found, skip.
            2. If 0 matches, search from VIAF and if 1 result found, create a new authority record from the data.
            3. If none or more than one responses found, use only Classificator data (coming from Linker?).

        for SIERRA records: normalize.
        """
        linked_records = []

        for linked in linking_results:
            entity = linked.get("original_entity")
            try:
                linked_info = linked.get("linked_info", [])
                linked_num = len(linked_info)

                if not linked_info:
                    # new record will be created
                    logger.info(
                        f"No linked entities found for {entity}")
                    continue

                if linked_num == 1:
                    linked = linked_info[0]
                    linked_records.append(linked.get("json", {}))
                    self.records_extra_data.append({
                        "entity": entity,
                        "viaf": linked.get("viaf", {}),
                        "type": "linked",
                        "edited": True
                    })
                else:
                    # new record will be created
                    logger.info(
                        f"Multiple linked entities found for {entity}")

            except Exception as e:
                logger.error(f"Error processing entity {entity}: {e}")

        self.records_extra_data.extend(
            {
                "sierraID": obj.get("sierraID"),
                "type": "sierra",
                "edited": True
            }
            for obj in (sierra_data or [])
        )

        all_records = linked_records + (sierra_data or [])

        return SafeJSONReader(
            json.dumps(all_records, ensure_ascii=False),
        )

    def _setup_records(self, linking_results: List[dict], sierra_data: List[dict]) -> JSONReader:
        """Setup initial MARC records and data.

        For linked entities:
            1. Try to get single linked normalized record from KATA elastic. If more than one found, skip.
            2. If 0 matches, search from VIAF and if 1 result found, create a new authority record from the data.
            3. If none or more than one responses found, use only Classificator data (coming from Linker?).
        """
        linked_records = []

        for linked in linking_results or []:
            if not isinstance(linked, dict):
                continue

            entity = linked.get("original_entity")
            linked_info = linked.get("linked_info", [])

            if not isinstance(linked_info, list) or not linked_info:
                continue

            if len(linked_info) == 1:
                linked_item = linked_info[0]
                if not isinstance(linked_item, dict):
                    continue

                linked_records.append(linked_item.get("json", {}))
                self.records_extra_data.append({
                    "entity": entity,
                    "viaf": linked_item.get("viaf", {}),
                    "type": "linked",
                    "edited": True
                })

        self.records_extra_data.extend(
            {
                "sierraID": obj.get("sierraID"),
                "type": "sierra",
                "edited": True
            }
            for obj in (sierra_data or [])
            if isinstance(obj, dict)
        )

        all_records = linked_records + (sierra_data or [])

        return SafeJSONReader(json.dumps(all_records, ensure_ascii=False))

    @staticmethod
    def current_timestamp():
        """6 digit timestamp."""
        return datetime.now().strftime("%H%M%S")

    @staticmethod
    def current_yyyy_dd():
        """format of 2025-03"""
        return datetime.now().strftime("%Y-%m")

    @staticmethod
    def _is_person_est_nationality(viaf_record: VIAFRecord) -> bool:
        return viaf_record.nationality == "ee"

    def _is_nxx(self, field: Field, n: str):
        """ Check if fields tag is in nxx range. """
        return field.tag.startswith(n)

    def get_record_field_or_none(self, record: Record, tag: str) -> Optional[Field]:
        return record.get_fields(tag)[0] if record.get_fields(tag) else None

    def _field_in_record(self, field: Field, record: Record) -> bool:
        """ Check if field exists in record. """
        existing_fields = record.get_fields(field.tag)
        return any(
            field.data == existing_field.data for existing_field in existing_fields)

    def _filter_equivalent_field_not_in_record(self, record: Record, fields: List[Field]) -> bool:
        """ filter out fields, that do not have an equivalent in the record. """
        return filter(lambda field: not self._field_in_record(field, record), fields)

    def get_subfield(self, record: Record, tag: str, subfield: str, default: str) -> str:
        """ get record existing subfield value or assign a fallback value. """

        field = self.get_record_field_or_none(record, tag)

        if field is None:
            return default

        subfields = field.get_subfields(subfield)
        return subfields[0] if subfields else default

    def _handle_default_fields(self, record: Record, *fields: List[Field]) -> Record:
        """ add field to record iff not present already """
        record.add_field(
            *filter(lambda field: field.tag not in [
                f.tag for f in record.get_fields()], fields)
        )

    def _handle_editable_fields(self, record: Record, *fields: List[Field]) -> Record:
        """ replace existing field with a new field. """

        editable_fields = filter(
            lambda field: field.tag in self.ALLOW_EDIT_FIELDS, fields)

        tags = [f.tag for f in editable_fields]

        record.remove_fields(
            *tags
        )
        record.add_field(
            *editable_fields
        )

    def _handle_repeatable_fields(self, record: Record, *fields: List[Field]) -> Record:
        """ add field to the record & don't replace existing field."""

        repeatable_fields = [
            field for field in fields if field.tag in self.REPEATABLE_FIELDS]

        record.add_field(
            *repeatable_fields
            # *self._filter_equivalent_field_not_in_record(
            #     record, repeatable_fields)
        )

    def _add_fields_to_record(self, record: Record, fields: List[Field]) -> Record:

        self._handle_repeatable_fields(record, *fields)
        self._handle_editable_fields(record, *fields)
        self._handle_default_fields(record, *fields)

    def _add_author(self, record: Record, viaf_record: VIAFRecord) -> Optional[Field]:

        existing_author: Optional[Field] = record.get(
            "100") or record.get("110") or record.get("111")
        if existing_author:
            return record

        type_map = {
            "Personal": "100",
            "Corporate": "110",
            "Collective": "111"
        }

        author_type = viaf_record.name_type
        tag = type_map.get(author_type, "100")

        fields = [
            Field(
                tag=tag,
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("a", viaf_record.name),
                    Subfield("b", viaf_record.name_type), # Is this correct??
                    Subfield("c", viaf_record.name_type)  # Is this correct??
                ]
            )
        ]

        self._add_fields_to_record(record, fields)

    def _normalize_common(self, record: Record, is_editing_existing_record: bool) -> None:
        """Common logic for all normalizations. """

        note = "Muudetud AI poolt" if is_editing_existing_record else "Loodud AI poolt"
        date_note = f"KRATT {self.current_yyyy_dd()}"

        fields = [
            Field(tag="667",
                  indicators=EMPTY_INDICATORS,
                  subfields=[Subfield("a", note)]),
            Field(tag="925",
                  indicators=EMPTY_INDICATORS,
                  subfields=[Subfield("t", self.get_subfield(record, "925", "t", date_note))
                             ] + ([Subfield("p", self.get_subfield(record, "925", "p", date_note))]
                                  if is_editing_existing_record else []))
        ]

        self._add_fields_to_record(record, fields)

        return record

    def _get_viaf_search_term(self, record: Record, entity: Optional[str]) -> Optional[str]:
        """ prioritize entity name, if not available, use author name. """
        if entity:
            return entity

        author_field = record.get("100") or record.get("110") or record.get("111")
        if author_field:
            return author_field.get_subfields("a")[0] if author_field.get_subfields("a") else None

        logger.warning(
            "No entity or author name found for VIAF search. Skipping VIAF enrichment.")

    def _get_viaf_record(self, record: Record, viaf_id: Optional[int] = None,
            entity: Optional[str] = None, viaf_field: str = DEFAULT_VIAF_FIELD,
            threshold: float = VIAF_SIMILARITY_THRESHOLD, verify: bool = VERIFY_VIAF_RECORD,
            max_records: int = MAX_VIAF_RECORDS_TO_VERIFY
    ) -> Optional[VIAFRecord]:
        viaf_record = None

        try:
            viaf_client = VIAFClient()

            if viaf_id:
                viaf_records = viaf_client.get_normalized_data_by_ids([viaf_id])
                if viaf_records:
                    viaf_record = viaf_records[0]
            else:
                search_term = self._get_viaf_search_term(record, entity)

                if search_term:
                    logger.info(
                        f"Searching for VIAF record with search term: {search_term}")

                    if not verify:
                        logger.warning(
                            f"Record verification is turned off. If multiple records are " \
                            f"detected for search term '{search_term}', the first " \
                            f"result is automatically returned. This might lead to " \
                            f"some inaccuracies!"
                        )

                    viaf_record = viaf_client.get_normalized_data_by_search_term(
                        search_term=search_term,
                        field=viaf_field,
                        max_records=max_records,
                        verify=verify,
                        threshold=threshold
                    )

        except Exception as e:
            logger.error(
                f"Error fetching VIAF record with ID={viaf_id} / entity='{entity}': {e}"
            )
        return viaf_record

    def _normalize_record(self, record: Record, sierraID: str,
                          viaf_record: VIAFRecord, is_editing_existing_record: bool) -> Record:
        return record

    @property
    def data(self) -> List[dict]:
        """Shorthand to get all normalized records as dict, skipping failures."""
        result = []
        for record in self:
            try:
                result.append(record.as_dict())
            except Exception as e:
                logger.error(f"Failed to normalize record: {e}")
                continue
        return result

    def __iter__(self) -> Iterator:
        viaf_id_path = "viaf.queryResult.records.record.0.recordData.VIAFCluster.viafID"
        sierra_id_path = "sierraID"

        for record, extra_data in zip(self.records, self.records_extra_data):

            sierra_id = glom(extra_data, sierra_id_path, default="")
            viaf_id = glom(extra_data, viaf_id_path, default=None)

            entity = extra_data.get("entity")
            is_editing_existing_record = extra_data.get("edited") == True

            viaf_record = self._get_viaf_record(record, viaf_id, entity)
            record = self._normalize_common(record, is_editing_existing_record)

            normalized_record = self._normalize_record(
                record, sierra_id, viaf_record, is_editing_existing_record)

            normalized_record.fields.sort(key=lambda field: field.tag)

            yield normalized_record



class Status:
    SKIPPED = "SKIPPED"
    FAILED = "FAILED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    BLOCKED = "BLOCKED"
    RETRYING = "RETRYING"


class Queue:
    CORE = "core"
    EMS = "ems"
    SIERRA = "sierra"


class Tasks:
    SEND_VERSION = "send_version_to_core"
    UPDATE_TASK_STATUS = "update_task_status"
    UPDATE_TASK_VALUES = "update_task_values"
    MODEL_UPDATE = "component_model_update"
    RUN_POST_TASK_COMPLETION_TASKS = "run_post_task_completion_tasks"


class Models:
    IMAGE_CLASSIFIER = "Pildid"

    TOPIC_KEYWORDS_ET = "Teemamärksõnad (ET)"
    TOPIC_KEYWORDS_EN = "Teemamärksõnad (EN)"

    FORM_KEYWORDS_ET = "Vormimärksõnad (ET)"
    FORM_KEYWORDS_EN = "Vormimärksõnad (EN)"

    TIME_KEYWORDS_ET = "Ajamärksõnad (ET)"
    TIME_KEYWORDS_EN = "Ajamärksõnad (EN)"

    UDK_ET = "UDK (ET)"
    UDK_EN = "UDK (EN)"
    UDC_ET = "UDC (ET)"
    UDC_EN = "UDC (EN)"

    DOMAIN_KEYWORDS_ET = "Märksõnade valdkonnad (ET)"
    DOMAIN_KEYWORDS_EN = "Märksõnade valdkonnad (EN)"



from pymarc import Record, Field, Subfield, Leader, JSONReader
import logging

logger = logging.getLogger(__name__)

DEFAULT_LEADER = '01682nz a2200349n 4500'

class SafeJSONReader(JSONReader):

    def __next__(self):
        while True:
            try:
                jobj = next(self.iter)
                rec = Record()

                # Use custom default leader if missing
                leader_str = jobj.get("leader")
                if leader_str:
                    rec.leader = Leader(leader_str)
                else:
                    logger.warning("Missing leader in record. Using DEFAULT_LEADER.")
                    rec.leader = Leader(DEFAULT_LEADER)

                for field in jobj["fields"]:
                    k, v = list(field.items())[0]

                    if isinstance(v, dict) and "subfields" in v:
                        subfields = []
                        for sub in v["subfields"]:
                            for code, value in sub.items():
                                subfields.append(Subfield(code, value))
                        ind1 = v.get("ind1", " ")
                        ind2 = v.get("ind2", " ")
                        fld = Field(tag=k, indicators=[ind1, ind2], subfields=subfields)
                    else:
                        fld = Field(tag=k, data=v)
                    rec.add_field(fld)

                return rec

            except StopIteration:
                raise
            except Exception as e:
                logger.error(f"Skipping invalid record: {e}")
                continue



COMPONENT_KEY = "meta_extractor"


class Tasks:
    SINGLE = "extract_meta_from_text"
    PIPELINE = "run_meta_extractor_with_core_logic"


class Queue:
    MAIN = "meta_extractor"


class StatusKeys:
    EXTRACT_METADATA = "extract_metadata"


import pytest
import os
from pprint import pprint
from rara_tools.core_formatters.core_formatter import format_keywords, format_meta
from tests.test_utils import read_json_file

ROOT_DIR = os.path.join("tests", "test_data", "formatter")
INPUT_KEYWORDS_FILE_PATHS = [
    os.path.join(ROOT_DIR, "keywords_1.json"),
    os.path.join(ROOT_DIR, "keywords_2.json")
]
INPUT_META_FILE_PATHS = [
    os.path.join(ROOT_DIR, "epub_meta.json"),
    os.path.join(ROOT_DIR, "mets_alto_meta.json"),
    os.path.join(ROOT_DIR, "pdf_meta.json")
]

INPUT_KEYWORDS = [
    read_json_file(keyword_file_path)
    for keyword_file_path in INPUT_KEYWORDS_FILE_PATHS
]

INPUT_META_DICTS = [
    read_json_file(meta_file_path)
    for meta_file_path in INPUT_META_FILE_PATHS
]

def test_formatting_keywords_for_core():
    for keyword_dict in INPUT_KEYWORDS:
        formatted_keywords = format_keywords(keyword_dict)
        #pprint(formatted_keywords)
        assert formatted_keywords
        assert isinstance(formatted_keywords, list)


def test_formatting_meta_for_core():
    for meta_dict in INPUT_META_DICTS:
        formatted_meta = format_meta(meta_dict)
        #pprint(formatted_meta)
        assert formatted_meta
        assert isinstance(formatted_meta, dict)



from rara_tools.constants.normalizers import EntityType

COMPONENT_KEY = "subject_indexer"


class Tasks:
    SINGLE = "run_subject_indexer_process"
    PIPELINE = "run_subject_indexer_with_core_logic"


class Queue:
    MAIN = "subject-indexer"


class StatusKeys:
    EXTRACT_KEYWORDS = "extract_keywords"


class URLSource:
    VIAF = "VIAF"
    SIERRA = "Sierra"
    EMS = "EMS"

class KeywordType:
    LOC = "Kohamärksõnad"
    TIME = "Ajamärksõnad"
    TOPIC = "Teemamärksõnad"
    GENRE = "Vormimärksõnad"
    TITLE = "Teose pealkiri"
    PER = "Isikunimi"
    ORG = "Kollektiivi nimi"
    EVENT = "Ajutine kollektiiv või sündmus"
    CATEGORY = "Valdkonnamärksõnad"
    UDC = "UDC Summary"
    UDK = "UDK Rahvusbibliograafia"


class KeywordMARC:
    PER = 600
    ORG = 610
    TOPIC = 650
    GENRE = 655
    TIME = 648
    LOC = 651
    EVENT = 611
    TITLE = 630

class KeywordSource:
    EMS = "EMS"
    SIERRA = "SIERRA"
    VIAF = "VIAF"
    AI = "AI"


KEYWORD_TYPE_MAP = {
    KeywordType.TIME: EntityType.KEYWORD,
    KeywordType.GENRE: EntityType.KEYWORD,
    KeywordType.LOC: EntityType.LOC,
    KeywordType.PER: EntityType.PER,
    KeywordType.ORG: EntityType.ORG,
    KeywordType.TOPIC: EntityType.KEYWORD,
    KeywordType.TITLE: EntityType.TITLE,
    KeywordType.EVENT: EntityType.ORG
}

KEYWORD_MARC_MAP = {
    KeywordType.LOC: KeywordMARC.LOC,
    KeywordType.TIME: KeywordMARC.TIME,
    KeywordType.TOPIC: KeywordMARC.TOPIC,
    KeywordType.GENRE: KeywordMARC.GENRE,
    KeywordType.TITLE: KeywordMARC.TITLE,
    KeywordType.ORG: KeywordMARC.ORG,
    KeywordType.PER: KeywordMARC.PER,
    KeywordType.EVENT: KeywordMARC.EVENT
}

KEYWORD_TYPES_TO_IGNORE = [
    KeywordType.CATEGORY,
    KeywordType.UDC,
    KeywordType.UDK
]

EMS_ENTITY_TYPES = [EntityType.KEYWORD, EntityType.LOC]
SIERRA_ENTITY_TYPES = [EntityType.PER, EntityType.ORG, EntityType.TITLE]
VIAF_ENTITY_TYPES = [EntityType.PER, EntityType.ORG, EntityType.TITLE]



import requests
import json
import regex as re
from typing import List, Dict
from collections import defaultdict
from jellyfish import jaro_winkler_similarity as jw
from requests.models import Response
from rara_tools.parsers.tools.entity_normalizers import PersonalName, Normalizer
from rara_tools.constants.normalizers import (
    DEFAULT_VIAF_FIELD, ALLOWED_VIAF_FIELDS, ALLOWED_VIAF_WIKILINK_LANGS,
    VIAF_SIMILARITY_THRESHOLD, VIAF_ALLOWED_SOURCES
)

import logging
logger = logging.getLogger(__name__)

class VIAFRecord:
    """ Takes in a VIAF query response JSON and wraps
    information extraction from it.
    """
    def __init__(self,
            record: dict,
            allowed_sources: List[str] = VIAF_ALLOWED_SOURCES
    ):
        """ Initializes VIAFRecord class.

        Parameters
        -----------
        record: dict
            VIAF query response JSON.
        allowed_sources: List[str]
            Only exracts information from these sources. Other
            sources are ignored.
        """
        self.__record: dict = record
        self.__record_data: dict = {}
        self.__allowed_sources: List[str] = allowed_sources
        self.__viaf_id: int = None
        self.__viaf_url: str = ""
        self.__name_variations: List[str] = []
        self.__birth_date: str = None
        self.__death_date: str = None
        self.__occupations: List[str] = []
        self.__all_fields: dict = {}
        self.__nationality: str = ""
        self.__has_isni: bool = False
        self.__name: str = ""
        self.__name_type: str = ""
        self.__has_isni: str = ""
        self.__activity_start: str = None
        self.__activity_end: str = None,
        self.__works: List[str] = []
        self.__wikilinks: dict = {}
        self.__all_wikilinks: List[str] = []
        self.__has_isni: bool | None = None
        self.__marc_400: List[dict] = []
        self.__marc_500: List[dict] = []
        self.__marc_main: List[dict] = []
        self.__subfield_indicator: str = ""

        self.__value_fields: List[str] = [
            "text", "value", "title", "datafield"
        ]
        self.__title_types: List[str] = ["UniformTitleWork"]


    def __get_data(self, field_name: str, subfield_name: str = "data",
            allowed_sources: List[str] = []
    ) -> List[str]:

        if not allowed_sources:
            allowed_sources = self.__allowed_sources

        data = []

        try:
            entries = self.record_data.get(
                field_name, {}
                ).get(subfield_name, [])

            for entry in entries:
                sources = entry.get("sources", {}).get("s", [])
                if set(allowed_sources).intersection(set(sources)):
                    for field in self.__value_fields:
                        value = entry.get(field, "")
                        if value:
                            data.append(value)
                            break
        except Exception as e:
            logger.error(
                f"Failed extracting data from field '{field_name}' with subfield " \
                f"'{subfield_name}'. '{field_name}' dict has the following " \
                f"structure: {self.record_data.get(field_name)}. " \
                f"Exception reason: {e}."
            )
        return data

    def _get_wikilink_lang(self, wikilink: str) -> str:
        """ Parses the language of the Wikipedia page
        from wikilink.
        """
        pattern = r"(?<=https\W{3})\w+(?=[.])"
        match = re.search(pattern, wikilink)
        wikilink_lang = ""
        if match:
            wikilink_lang = match.group()
        return wikilink_lang

    def _get_marc_field(self, marc_dict: dict, subfield: str = "a",
                        strict_subfield: bool = True
    ) -> str:
        """ Retrieve value from a MARC dict

        Parameters
        -----------
        marc_dict: dict
            MARC dictionaryself.
        subfield: str
            Subfield to extract
        strict_subfield: bool
            If set to True, data is extracted ONLY from
            the subfield set with param `subfield`. If set to False,
            data can be extracted from other subfields as well as long
            there is only one subfield in the dict. This might be necessary
            for uniformTitleWorks as sometimes the title is present in
            subfield (t) while subfield (a) contains the author. However,
            there are instances, where the title is present is subfield (a)
            with no author.
        """
        value = ""
        if marc_dict.get("dtype", "") == "MARC21":
            subfields = marc_dict.get("subfield", [])
            for _subfield in subfields:
                if len(subfields) > 1 and _subfield.get("code", "") == subfield:
                    value = _subfield.get("value", "")
                    break
                elif len(subfields) == 1 and not strict_subfield:
                    value = _subfield.get("value", "")
        return value

    def _get_marc_tag(self, marc_dict: dict) -> str:
        tag = ""
        if marc_dict.get("dtype", "") == "MARC21":
            tag = marc_dict.get("tag", "")
        return tag

    def _get_names(self, marc_dicts: List[dict]) -> List[str]:
        names_d = defaultdict(int)
        for marc_dict in marc_dicts:
            name = self._get_marc_field(
                marc_dict=marc_dict,
                subfield=self.subfield_indicator,
                strict_subfield=False
            )
            names_d[name]+=1
        name_list = sorted(
            list(names_d.items()),
            key=lambda x: x[1],
            reverse=True
        )
        names = []
        for n in name_list:
            _name = self._strip_punctuation(n[0])
            if _name not in names:
                names.append(_name)

        return names

    def _get_name(self, marc_dicts: List[dict]) -> str:
        names = self._get_names(marc_dicts)
        name = ""
        if names:
            name = names[0]
        return name

    def _strip_punctuation(self, entity: str) -> str:
        entity = entity.strip(",")
        # Strip "." only if the last token is not an initial,
        # e.g: "Meri, Lennart." -> Strip
        # "Meri, L." -> Do not strip.
        ent_tokens = [t.strip() for t in entity.split() if t.strip()]
        if ent_tokens and len(ent_tokens[-1]) > 2:
            entity = entity.strip(".")
        return entity

    def _strip_parenthesis(self, entity: str) -> str:
        """ Strip information in parenthesis from VIAF records
        in order to compare the records more easily.
        """
        _entity = re.sub(r"[(][^)][)]", "", entity)
        return _entity.strip()

    @property
    def record(self) -> dict:
        return self.__record

    @property
    def record_data(self) -> dict:
        return self.__record_data

    @property
    def subfield_indicator(self) -> str:
        if not self.__subfield_indicator:
            if self.name_type in self.__title_types:
                subfield_name = "t"
            else:
                subfield_name = "a"
            self.__subfield_indicator = subfield_name
        return self.__subfield_indicator

    @property
    def name(self) -> str:
        # author -> name
        if not self.__name:
            if self.marc_main:
                self.__name = self._get_name(self.marc_main)
            else:
                names = self.__get_data("mainHeadings", "data")
                if names:
                    self.__name = names[0]
        return self.__name

    @property
    def name_type(self) -> str:
        # author_type -> name_type
        """ Type of name (personal, corporate, title, etc)
        """
        if not self.__name_type:
            self.__name_type = self.record_data.get("nameType")
        return self.__name_type

    @property
    def viaf_id(self) -> int:
        if not self.__viaf_id:
            self.__viaf_id = self.record_data.get("viafID", "")
        return self.__viaf_id

    @property
    def viaf_url(self) -> str:
        if not self.__viaf_url:
            self.__viaf_url = self.record_data.get(
                "Document", {}).get("about", "")
        return self.__viaf_url

    @property
    def has_isni(self) -> bool:
        if self.__has_isni == None:
            self.__has_isni = bool(self.record_data.get("isni", ""))
        return self.__has_isni

    @property
    def record_data(self) -> dict:
        if not self.__record_data:
            try:
                self.__record_data = self.__record["queryResult"]
            except:
                self.__record_data = self.__record["recordData"]["VIAFCluster"]

        return self.__record_data

    @property
    def name_variations(self) -> List[str]:
        if not self.__name_variations:
            if self.marc_400:
                var_1 = self._get_names(self.marc_400)
                var_2 = self._get_names(self.marc_main)
                _vars = var_1 + var_2

            else:
                _vars = self.__get_data("mainHeadings")
            vars_3 = [Normalizer.clean_entity(v) for v in _vars]

            vars = _vars + vars_3
            self.__name_variations = list(set(vars))
        return self.__name_variations

    @property
    def birth_date(self) -> str:
        if not self.__birth_date:
            self.__birth_date = self.record_data.get("birthDate", None)
        return self.__birth_date

    @property
    def death_date(self) -> str:
        if not self.__death_date:
            self.__death_date = self.record_data.get("deathDate", None)
        return self.__death_date

    @property
    def occupations(self) -> List[str]:
        if not self.__occupations:
            self.__occupations = self.__get_data("occupation")
        return self.__occupations

    @property
    def activity_start(self) -> str:
        if not self.__birth_date:
            self.__birth_date = self.record_data.get("activityStart", None)
        return self.__birth_date

    @property
    def activity_end(self) -> str:
        if not self.__death_date:
            self.__death_date = self.record_data.get("activityEnd", None)
        return self.__death_date

    @property
    def nationality(self) -> str:
        if not self.__nationality:
            nationalities = self.__get_data("nationalityOfEntity")
            nationalities_dict = defaultdict(int)
            for n in nationalities:
                nationalities_dict[n.lower()] += 1
            if nationalities:
                self.__nationality = sorted(
                    nationalities_dict.items(),
                    key=lambda x: x[1],
                    reverse=True
                )[0][0]
        return self.__nationality

    @property
    def works(self) -> List[str]:
        if not self.__works:
            self.__works = list(set(self.__get_data(
                field_name="titles",
                subfield_name="work"
            )))
        return self.__works

    @property
    def all_wikilinks(self) -> List[str]:
        if not self.__all_wikilinks:
            self.__all_wikilinks = self.__get_data(
                field_name="xLinks", subfield_name="xLink",
                allowed_sources=["WKP"]
            )
        return self.__all_wikilinks

    @property
    def wikilinks(self) -> dict:
        if not self.__wikilinks:
            for wikilink in self.all_wikilinks:
                wikilink_lang = self._get_wikilink_lang(wikilink)
                if wikilink_lang and wikilink_lang in ALLOWED_VIAF_WIKILINK_LANGS:
                    self.__wikilinks[wikilink_lang] = wikilink
        return self.__wikilinks

    @property
    def marc_400(self) -> List[dict]:
        if not self.__marc_400:
            self.__marc_400 = self.__get_data(
                field_name="x400s",
                subfield_name="x400"
            )
        return self.__marc_400

    @property
    def marc_500(self) -> List[dict]:
        if not self.__marc_500:
            self.__marc_500 = self.__get_data(
                field_name="x500s",
                subfield_name="x500"
            )
        return self.__marc_500


    @property
    def marc_main(self) -> List[dict]:
        if not self.__marc_main:
            self.__marc_main = self.__get_data(
                field_name="mainHeadings",
                subfield_name="mainHeadingEl"
            )
        return self.__marc_main

    @property
    def all_fields(self) -> dict:
        if not self.__all_fields:
            self.__all_fields = {
                "viaf_id": self.viaf_id,
                "viaf_url": self.viaf_url,
                "name": self.name,
                "name_type": self.name_type,
                "name_variations": self.name_variations,
                "birth_date": self.birth_date,
                "death_date": self.death_date,
                "occupations": self.occupations,
                "nationality": self.nationality,
                "activity_start": self.activity_start,
                "activity_end": self.activity_end,
                "has_isni": self.has_isni,
                "works": self.works,
                "wikilinks": self.wikilinks,
                "marc_400": self.marc_400,
                "marc_500": self.marc_500,
                "marc_main": self.marc_main
            }
        return self.__all_fields


class VIAFClient:
    def __init__(self,
            viaf_api_url: str = "https://viaf.org/api",
            allowed_viaf_sources: List[str] = VIAF_ALLOWED_SOURCES
        ):
        self.root_url: str = viaf_api_url.strip("/")
        self.record_url: str = f"{self.root_url}/cluster-record"
        self.search_url: str = f"{self.root_url}/search"
        self.headers: dict = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        self.allowed_viaf_sources: List[str] = allowed_viaf_sources

    def check_search_term_query(self) -> bool:
        """ Function for checking, if VIAF search term
        query works as expected.
        """
        test_entity = "Lennart Meri"
        record = self.get_normalized_data_by_search_term(
            search_term=test_entity,
            max_records=1,
            verify=False
        )
        success = True

        if record:
            if record.name != "Meri, Lennart":
                success = False
        else:
            success = False
        if not success:
            logger.error(f"VIAF search term query has changed or not working!")
        return success

    def check_id_query(self) -> bool:
        """ Function for checking, if VIAF search term
        query works as expected.
        """
        test_id = "84153775"
        records = self.get_normalized_data_by_ids([test_id])
        success = True
        if records:
            record = records[0]
            if record.name != "Meri, Lennart":
                success = False
        else:
            success = False

        if not success:
            logger.error(f"VIAF ID query has changed or not working!")
        return success

    @staticmethod
    def verify(entity: str, viaf_record: VIAFRecord,
            threshold: float = VIAF_SIMILARITY_THRESHOLD
    ) -> dict:
        """ Verifies, if entity to link is sufficiently
        similar to a VIAF Record based on name forms in
        VIAFRecord.name_variations.

        Parameters
        ------------
        entity: str
            Entity queried from VIAF.
        viaf_record: VIAFRecord
            A VIAFRecord object.
        threshold: float
            Min similarity threshold for a verified result
            Should be a float between 0 and 1.

        Returns
        ------------
        dict
            Dict with keys:
                verified: bool
                    If the VIAFRecord was verified to be
                    sufficiently similar.
                most_similar_record: str
                    The most similar string to entity
                    in VIAFRecord.name_variations.
                score: float
                    Similarity score of the most similar record.
        """
        logger.debug(
            f"Verifying if '{viaf_record.name}' is sufficiently similar to '{entity}'."
        )
        # might not always be personal name, but shouldn't break anything
        if len(entity.split()) > 1:
            pn = PersonalName(entity)
            name_forms = [pn.last_comma_first, pn.first_last]
        else:
            name_forms = [entity]
        max_similarity = 0
        most_similar_record = ""
        verified = False
        for var in viaf_record.name_variations:
            for name_form in name_forms:
                score = jw(name_form.lower(), var.lower())
                if score > max_similarity:
                    max_similarity = score
                    most_similar_record = var
                if score >= threshold:
                    logger.debug(
                        f"Verification successful! '{name_form}' sufficiently " \
                        f"similar to '{var}'! Score = {score}."
                    )
                    verified = True
                    break
            if verified:
                break
        out = {
            "verified": verified,
            "most_similar_record": most_similar_record,
            "score": max_similarity
        }
        return out

    @staticmethod
    def get_verified_record(search_term: str, viaf_records: List[VIAFRecord],
        threshold: float = VIAF_SIMILARITY_THRESHOLD
    ) -> VIAFRecord:
        """ Takes in n VIAFRecords found while searching the term `search_term`.
        Returns the most similar VIAFRecord.
        """
        logger.debug(
            f"Retrieving a single verified record from VIAF search results. " \
            f"search term = '{search_term}'."
        )
        verified_record = None
        max_score = 0
        most_similar_record = ""
        for record in viaf_records:
            verified = VIAFClient.verify(search_term, record, threshold)
            if verified.get("score") > max_score:
                most_similar_record = verified.get("most_similar_record")
                max_score = verified.get("score")
            if verified.get("verified"):
                verified_record = record
                break
        if not verified_record:
            logger.error(
                f"Verification failed. No matched record surpassed the set similarity " \
                f"threshold ({threshold}). Closest match for search term '{search_term}' was " \
                f"'{most_similar_record}' with similarity score {max_score} "
            )
        return verified_record

    def _send_request(self, url: str, data: dict) -> Response:
        return requests.post(url, data=json.dumps(data), headers=self.headers)

    def get_records_by_search_term(self,
            search_term: str,
            index: str = "VIAF",
            field: str = DEFAULT_VIAF_FIELD,
            page_index: int = 0,
            page_size: int = 50
     ) -> Response:
        """ Query VIAF records by search term.
        """
        logger.debug(f"Retriecing VIAF records for search term '{search_term}'.")
        if field and field not in ALLOWED_VIAF_FIELDS:
            logger.error(
                f"Field '{field}' is not allowed. Defaulting to '{DEFAULT_VIAF_FIELD}'. " \
                f"Allowed VIAF fields are: {ALLOWED_VIAF_FIELDS}. "
            )
            field = DEFAULT_VIAF_FIELD
        data = {
            "reqValues": {
                "field": field,
                "index": index,
                "searchTerms": search_term
            },
            "meta": {
                "env": "prod",
                "pageIndex": page_index,
                "pageSize": page_size
            }
        }
        response = self._send_request(url=self.search_url, data=data)
        return response

    def get_records_by_viaf_id(self, record_id: str) -> Response:
        """ Query VIAF records by ID.
        """
        logger.debug(f"Retrieving VIAF records for ID {record_id}.")
        data = {
            "reqValues": {
                "recordId": str(record_id)
            }
        }
        response = self._send_request(url=self.record_url, data=data)
        return response

    def extract_viaf_ids(self, search_query_response: Response) -> List[str]:
        """ Parse VIAF ID-s from search query response.
        """
        logger.debug("Extracting VIAF IDs from VIAF search query results.")
        try:
            records = search_query_response.json()["queryResult"]["records"]["record"]
        except Exception as e:
            logger.error(
                f"Parsing records from search query " \
                f"response failed with error: {e}."
            )
            records = []
        viaf_ids = []
        for record in records:
            try:
                viaf_id = record["recordData"]["VIAFCluster"]["viafID"]
                viaf_ids.append(viaf_id)
            except Exception as e:
                logger.error(
                    f"Extracing VIAF ID from record '{record}' " \
                    f"failed with error: {e}"
                )
        return viaf_ids

    def get_viaf_ids_by_search_terms(self,
            search_term: str, field: str = DEFAULT_VIAF_FIELD,
            viaf_index: str = "VIAF", page_size: int = 50
    ) -> List[str]:
        """ Get all matching VIAF IDs for a search term.
        """

        search_response = self.get_records_by_search_term(
            search_term=search_term,
            field=field,
            index=viaf_index,
            page_size=page_size
        )
        viaf_ids = self.extract_viaf_ids(search_response)
        return viaf_ids


    def fetch_viaf_clusters(self, viaf_ids: List[str]) -> Dict[str, dict]:
        results = {}
        for viaf_id in viaf_ids:
            try:
                response = self.get_records_by_viaf_id(viaf_id)
                response.raise_for_status()
                results[viaf_id] = response.json()
            except requests.exceptions.RequestException as e:
                logger.error(f"Error fetching VIAF record {viaf_id}: {e}")
                results[viaf_id] = {}

        return results

    def get_normalized_data_by_ids(self, record_ids: List[str]) -> List[VIAFRecord]:
        """ Fetch data required for normalization from VIAF. """
        logger.debug(f"Fetching VIAFRecords for the following IDs: {record_ids}.")
        response = self.fetch_viaf_clusters(record_ids)
        viaf_records = [
            VIAFRecord(
                record=response[record_id],
                allowed_sources=self.allowed_viaf_sources
            )
            for record_id in record_ids
        ]
        return viaf_records

    def get_normalized_data_by_search_term(self,
        search_term: str, field: str = DEFAULT_VIAF_FIELD, max_records: int = 10,
        verify: bool = True, threshold: float = VIAF_SIMILARITY_THRESHOLD,
        viaf_index: str = "VIAF"
    ) -> VIAFRecord | None:
        """ Fetch data required for normalization from VIAF. """
        logger.debug(
            f"Finding VIAFRecords with search term '{search_term}' " \
            f"using VIAF field='{field}', verify={verify}, threshold={threshold}. " \
            f"Allowed VIAF sources are: {self.allowed_viaf_sources}."
        )
        viaf_record = None
        verified_record = None
        viaf_ids = self.get_viaf_ids_by_search_terms(
            search_term=search_term,
            field=field,
            page_size=max_records,
            viaf_index=viaf_index
        )
        if verify:
            records = self.get_normalized_data_by_ids(viaf_ids[:max_records])
            verified_record = VIAFClient.get_verified_record(
                search_term=search_term,
                viaf_records=records,
                threshold=threshold
            )
        else:
            if viaf_ids:
                records = self.get_normalized_data_by_ids(viaf_ids[:1])
                verified_record = records[0] if records else None
        return verified_record



from rara_tools.constants import EMPTY_INDICATORS
from rara_tools.normalizers.viaf import VIAFRecord

from rara_tools.normalizers import RecordNormalizer

from pymarc import Field, Subfield, Record
from typing import List


class AuthoritiesRecordNormalizer(RecordNormalizer):
    """ Normalize authorities records """

    def __init__(self, linking_results: List[dict] = [], sierra_data: List[dict] = [],
                 ALLOW_EDIT_FIELDS: List[str] = [
                     "667", "925", "043"],
                 REPEATABLE_FIELDS: List[str] = ["024", "035", "400", "670"]):

        super().__init__(linking_results, sierra_data)
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS

    def _normalize_sierra(self, record: Record, sierraID: str) -> None:

        suffix_008 = "|n|adnnnaabn          || |a|      "

        fields = [
            Field(
                tag="008",
                indicators=EMPTY_INDICATORS,
                data=f"{self.current_timestamp()}{suffix_008}"
            ),

            Field(
                tag="040",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    # if record subfield exists already, use that value. if not, use hardcoded value
                    Subfield("a", self.get_subfield(
                        record, "040", "a", "ErESTER")),
                    Subfield("b", self.get_subfield(
                        record, "040", "b", "est")),
                    Subfield("c", self.get_subfield(
                        record, "040", "c", "ErEster")),
                ]
            ),
        ]

        self._add_fields_to_record(record, fields)

        return record

    def _add_birth_and_death_dates(self, record: Record, viaf_record: VIAFRecord) -> None:
        subfields_046 = [
            Subfield("f", self.get_subfield(
                record, "046", "f", viaf_record.birth_date)),
            Subfield("g", self.get_subfield(
                record, "046", "g", viaf_record.death_date)),
            Subfield("s", self.get_subfield(
                record, "046", "s", viaf_record.activity_start)),
            Subfield("t", self.get_subfield(
                record, "046", "t", viaf_record.activity_end)),
        ]

        self._add_fields_to_record(
            record, [Field(tag="046", indicators=EMPTY_INDICATORS, subfields=subfields_046)])

    def _add_viaf_url_and_isni(self, record: Record, viaf_record: VIAFRecord) -> None:
        # TODO 024. will be used to store KRATT KATA ID. Just generate one?
        viaf_url = f"https://viaf.org/viaf/{viaf_record.viaf_id}"

        subfields = [Subfield("0", self.get_subfield(
            record, "024", "0", viaf_url))]

        if viaf_record.has_isni:
            subfields.append(Subfield("2", "isni"))

        field = Field(tag="024", indicators=EMPTY_INDICATORS,
                      subfields=subfields)

        self._add_fields_to_record(record, [field])

    def _add_nationality(self, record: Record, viaf_record: VIAFRecord) -> None:

        fields = [
            Field(
                tag="043",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("c", "ee")
                ] if self._is_person_est_nationality(viaf_record) else []
            )]

        self._add_fields_to_record(record, fields)

    def _normalize_viaf(self, record: Record, viaf_record: VIAFRecord) -> None:
        """"
        Attempts to enrich the record with VIAF data.

        024 - repeatable field, add VIAF URL to subfield 0. If ISNI found, add to subfield 2
        043 - repeatable field. Add "ee" if found to be estonian nationality
        046 - non-repeatable field, add birth and death dates
        100, 110, 111 - non-repeatable field, attempts to add author type, if missing.

        """
        # TODO: include KRATT KATA ID to 024 and remove on delete. Increment last elastic ID?
        if not viaf_record:
            return

        self._add_nationality(record, viaf_record)
        self._add_viaf_url_and_isni(record, viaf_record)
        self._add_birth_and_death_dates(record, viaf_record)
        self._add_author(record, viaf_record)

    def _normalize_record(self, record: Record, sierraID: str,
                          viaf_record: VIAFRecord, is_editing_existing_record: bool) -> Record:

        self._normalize_sierra(record, sierraID)
        self._normalize_viaf(record, viaf_record)

        return record



from typing import List, NoReturn
from pymarc.record import Record
from rara_tools.parsers.marc_records.base_record import BaseRecord
from rara_tools.constants.parsers import TitleMarcIDs, LOGGER
import regex as re
import json
import logging


class TitleRecord(BaseRecord):
    """ Generates a simplified title JSON record
    from a pymarc MARC record.
    """
    def __init__(self, record: Record, add_variations: bool = False) -> NoReturn:
        """ Initializes TitleRecord object.

        Parameters
        -----------
        record: Record
            pymarc.record.Record object.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        super().__init__(record=record, add_variations=add_variations)

        self.__name_field_id: List[str] = TitleMarcIDs.NAME
        self.__name_variations_field_id: List[str]= TitleMarcIDs.NAME_VARIATIONS
        self.__year_field_id: List[str] = TitleMarcIDs.YEAR
        self.__type_field_id: List[str] = TitleMarcIDs.TYPE

        self.__default_year: int | None = None

        self.__name: str = ""
        self.__author_original_name: dict = {}
        self.__author_name: str = ""
        self.__year: int = -1
        self.__type: str = ""

        self.__author_life_years: str = ""
        self.__author_birth_year: int = -1
        self.__author_death_year: int = -1
        self.__name_variations: List[str] = []
        self.__full_record: dict = {}


    def _parse_year(self, year: str) -> int:
        year = year.strip()
        _year = self.__default_year
        if len(year) >= 4:
            if year[:4].isnumeric():
                _year = int(year[:4])
        elif len(year) == 3 and year.isnumeric():
            _year = int(year)
        return _year

    @property
    def name(self) -> str:
        if not self.__name:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id="t"
            )
            if values:
                self.__name = self._clean_value(values[0])
            else:
                pass
        return self.__name


    @property
    def year(self) -> str:
        if self.__year == -1:
            values = self.get_values(
                marc_ids=self.__year_field_id,
                subfield_id="k"
            )
            self.__year = self._clean_value(values[0]) if values else None
        return self.__year

    @property
    def type(self) -> str:
        if not self.__type:
            values = self.get_values(
                marc_ids = self.__type_field_id,
                subfield_id="a"
            )
            self.__type = self._clean_value(values[0]) if values else ""
        return self.__type

    @property
    def author_original_name(self) -> str:
        if not self.__author_original_name:
            values = self.get_values(
                marc_ids=self.__name_field_id,
                subfield_id=["a", "b"]
            )
            if values:
                self.__author_original_name = {
                    "a": self._clean_value(values[0].get("a", "")),
                    "b": self._clean_value(values[0].get("b", ""))
                }
            else:
                pass
        return self.__author_original_name

    @property
    def author_name(self) -> str:
        if not self.__author_name:
            self.__author_name = self._merge_and_clean(self.author_original_name, ["a", "b"])
        return self.__author_name

    @property
    def author_life_years(self) -> str:
        if not self.__author_life_years:
            values = self.get_values(
                marc_ids = self.__name_field_id,
                subfield_id="d"
            )
            self.__author_life_years = self._clean_value(values[0]) if values else ""
        return self.__author_life_years


    @property
    def author_birth_year(self) -> int:
        if self.__author_birth_year == -1:
            try:
                birth_year, death_year = self.author_life_years.split("-")
                self.__author_birth_year = self._parse_year(birth_year)
                self.__author_death_year = self._parse_year(death_year)
            except Exception as e:
                LOGGER.warning(
                    f"Failed extracting birth and/or death year " \
                    f"from '{self.author_life_years}' with the following " \
                    f"exception: '{e}'."
                )
        return self.__author_birth_year


    @property
    def author_death_year(self) -> int:
        if self.__author_death_year == -1:
            try:
                birth_year, death_year = self.author_life_years.split("-")
                self.__author_birth_year = self._parse_year(birth_year)
                self.__author_death_year = self._parse_year(death_year)
            except Exception as e:
                LOGGER.warning(
                    f"Failed extracting birth and/or death year " \
                    f"from '{self.author_life_years}' with the following " \
                    f"exception: '{e}'."
                )
        return self.__author_death_year

    @property
    def name_variations(self) -> List[str]:
        if not self.__name_variations:
            values = self.get_values(
                marc_ids=self.__name_variations_field_id,
                subfield_id="t"
            )
            variations = [self.name]
            if values:
                _variations = [
                    self._clean_value(value)
                    for value in values
                ]
                variations.extend(_variations)
            variations = [v.lower() for v in variations]
            self.__name_variations = list(set(variations))
        return self.__name_variations


    @property
    def full_record(self) -> dict:
        if not self.__full_record:
            self.__full_record = {
                "name": self.name,
                "author_name": self.author_name,
                "year": self.year,
                "type": self.type,
                "author_life_years": self.author_life_years,
                "author_birth_year": self.author_birth_year,
                "author_death_year": self.author_death_year,
                "name_variations": self.name_variations,
                "full_record_marc": str(self.marc_record),
                "full_record_json": json.dumps(self.marc_json_record)
            }
            if self.add_variations:
                self.__full_record.update(
                    {"link_variations": self.name_variations}
                )

        return self.__full_record



import os

import pytest
from rara_tools.s3 import S3Files

S3_FILES = S3Files(
    url = os.getenv("S3_URL", "s3.eu-central-003.backblazeb2.com"),
    access_key = os.getenv("S3_ACCESS_KEY", ""),
    secret_key = os.getenv("S3_SECRET_KEY", ""),
    bucket = os.getenv("S3_TEST_BUCKET", "rara-test")
)
TEST_PREFIX = "test/"
TEST_FILE = "./tests/test_data/test.txt"
TEST_FOLDER = "./tests/test_data/test_folder"
TEST_FILE_S3 = None
TEST_FOLDER_S3 = None


@pytest.mark.order(1)
def test_s3_listing():
    """Tests the file listing feature.
    """
    # test full_listing
    file_list = S3_FILES.list()
    assert len(file_list) > 1
    # test prefix listing
    file_list_eesti = S3_FILES.list(prefix="eesti")
    assert len(file_list_eesti) == 1
    file_list_mets_alto = S3_FILES.list(prefix="mets_alto")
    assert len(file_list_mets_alto) > 1

@pytest.mark.order(2)
def test_s3_file_upload():
    """Tests uploading a file to S3.
    """
    uploaded = S3_FILES.upload(TEST_FILE, prefix=TEST_PREFIX)
    assert uploaded in S3_FILES.list()
    # update global variable to reflect the actual file name given by the package
    global TEST_FILE_S3
    TEST_FILE_S3 = uploaded

@pytest.mark.order(3)
def test_s3_file_download():
    """Tests downloading a file from S3.
    """
    downloaded = list(S3_FILES.download(TEST_FILE_S3))
    assert f"./{TEST_FILE_S3}" in downloaded
    assert os.path.exists(TEST_FILE_S3)
    # remove downloaded files
    for f in downloaded:
        os.remove(f)
    # remove test dir
    os.rmdir("test")

@pytest.mark.order(4)
def test_s3_file_delete():
    """Tests deleting files from S3 bucket.
    """
    S3_FILES.delete(TEST_FILE_S3)
    assert TEST_FILE_S3 not in S3_FILES.list()

@pytest.mark.order(5)
def test_s3_folder_upload():
    """Tests uploading folders to S3 bucket.
    """
    uploaded = S3_FILES.upload(TEST_FOLDER, prefix=TEST_PREFIX)
    assert len(S3_FILES.list(prefix=uploaded)) == 3
    # update global variable to reflect the actual folder name given by the package
    global TEST_FOLDER_S3
    TEST_FOLDER_S3 = uploaded

@pytest.mark.order(6)
def test_s3_folder_download():
    """Tests downloading folders from S3 bucket.
    """
    downloaded = list(S3_FILES.download(TEST_FOLDER_S3))
    for dl_file in downloaded:
        assert dl_file.startswith(f"./{TEST_FOLDER_S3}")
    assert os.path.exists(TEST_FOLDER_S3)
    # remove downloaded files
    for f in downloaded:
        os.remove(f)
    # remove test dir
    os.rmdir(TEST_FOLDER_S3)
    os.rmdir("test")

@pytest.mark.order(7)
def test_s3_folder_delete():
    """Tests deleting folders from S3 bucket.
    """
    S3_FILES.delete(TEST_FOLDER_S3)
    assert len(S3_FILES.list(prefix=TEST_FOLDER_S3)) == 0
